{ File   : VDAVIEW4.PAS
  Author : Deniz Oezmen
  Created: ~2001
  Changed: 2005-08-26

  The Lost Files of Sherlock Holmes: The Case of The Serrated Scalpel VDA/VDX
  animation viewer.
  Note: Compile for protected mode.
}
program HolmesVDAViewer;
{$M 65521, 0, 655360}

uses
  VesaLib,
  Crt;

const
  MaxBlocks  = 4096;
  MaxSlices  = 512;
  EndOfSlice = $ffff;
  EndOfFile  = $fffe;
  FlagMask   = $f000;
  BlockMask  = $0fff;
  LogName    = 'VDAVIEW4.LOG';
  VDAExt     = '.VDA';
  VDXExt     = '.VDX';

type
  TColors    = array[0..255] of record
                                  R, G, B: Byte
                                end;

  TBlock     = record
                 PosX, PosY,
                 Width, Height: Word;
                 Content      : Pointer
               end;

  TFrame     = record
                 DataSize: Longint;
                 Content : Pointer
               end;


  TBlockList = array[0..MaxBlocks - 1] of TBlock;
                     (* ^^^^ wieviele Bits fr Blocknummern? 12? *)
                     (* evtl. auch 16 - 1 = 15?                  *)
  TFrameList = array[0..MaxSlices - 1] of TFrame;

var
  VDAFile,
  VDXFile    : file;
  Log        : Text;
  Colors     : TColors;
  BlockList  : TBlockList;
  ControlList: TFrameList;
  BlockCount,
  SliceCount,
  NrBlocks,
  NrSlices   : Word;

procedure ReadPalette(var Colors: TColors);
var
  i: Byte;
begin
  Writeln(Log, 'Reading VGA palette');
  BlockRead(VDAFile, Colors, SizeOf(Colors));

  Write(Log, 'Setting VGA palette ... ');
  for i := 0 to 255 do
    SetRGBPalette(i, Colors[i].R, Colors[i].G, Colors[i].B);

  Writeln(Log, 'done.')
end;

procedure ReadBlock(var Block: TBlock);
var
  b         : Byte;
  DataSize  : Longint;
begin
  with Block do
  begin
    PosX   := 0;
    PosY   := 0;
    Width  := 0;
    Height := 0;

    Write(Log, 'Reading block information ... ');
    if not Eof(VDAFile) then
    begin
      (* read block info *)
      BlockRead(VDAFile, Width, 2);
      BlockRead(VDAFile, Height, 1);
      Write(Log, 'size ', Width + 1, ' x ', Height + 1);

      (* jump over two unknown bytes (always zero?) *)
      Seek(VDAFile, FilePos(VDAFile) + 2);

      (* read default position *)
      BlockRead(VDAFile, PosX, 2);
      BlockRead(VDAFile, PosY, 1);
      Write(Log, ' at default position ', PosX, ', ', PosY, ' ... ');

      (* read block data and copy it to buffer *)
      DataSize := Longint(Height + 1) * Longint(Width + 1);
      GetMem(Content, DataSize);
      BlockRead(VDAFile, Content^, DataSize);
      Writeln(Log, 'done.')
    end
  end
end;

procedure PreloadBlocks(var BlockList: TBlockList; var NrBlocks: Word);
begin
  Writeln(Log, 'Preloading animation blocks');
  NrBlocks := 0;
  while not Eof(VDAFile) do
  begin
    Writeln(Log, 'Loading block #', NrBlocks);
    ReadBlock(BlockList[NrBlocks]);
    Inc(NrBlocks)
  end
end;

procedure DrawBlock(Block: TBlock);
var
  x, y: Word;
  p   : ^Byte;
begin
  with Block do
  begin
    p := Content;
    for y := PosY to PosY + Height do
      for x := PosX to PosX + Width do
      begin
        (* draw pixel in case it is not transparent *)
        if (p^ < 255) then
          PutPixel(x, y, p^);
        Inc(p)
      end
  end
end;

procedure DrawBlockAt(Block: TBlock; PosX, PosY: Word);
var
  x, y: Word;
  p   : ^Byte;
begin
  p := Block.Content;
  for y := PosY to PosY + Block.Height do
    for x := PosX to PosX + Block.Width do
    begin
      (* draw pixel in case it is not transparent *)
      if (p^ < 255) then
        PutPixel(x, y, p^);
      Inc(p)
    end
end;

procedure LoadControl(var ControlList: TFrameList; var NrSlices: Word);
var
  AnimBeginPos,
  AnimEndPos  : Longint;
  w           : Word;
  p           : ^Word;
begin
  Write(Log, 'Loading control file ... ');
  NrSlices := 0;

  while not Eof(VDXFile) do
  begin
    (* save start position of current slice *)
    AnimBeginPos := FilePos(VDXFile);

    (* search and save end position *)
    w := 0;
    while (w <> EndOfSlice) and not Eof(VDXFile) do
      BlockRead(VDXFile, w, SizeOf(w));
    AnimEndPos := FilePos(VDXFile);

    (* load data into buffer *)
    Seek(VDXFile, AnimBeginPos);
    with ControlList[NrSlices] do
    begin
      DataSize := AnimEndPos - AnimBeginPos;
      GetMem(Content, DataSize);
      BlockRead(VDXFile, Content^, DataSize);
    end;
    Inc(NrSlices)
  end;

  Writeln(Log, NrSlices, ' slices found.')
end;

procedure DrawSequence(ControlList: TFrameList; Slice: Word);
var
  ListPtr: ^Word;
  PosX,
  PosY,
  Command: Word;
begin
  Writeln(Log, 'Drawing sequence of slice #', Slice);

  ListPtr := ControlList[Slice].Content;
  Command := 0;

  (* proceed as long as end of slice has not been reached *)
  while (Command <> $ffff) do
  begin
    (* get next command *)
    Command := ListPtr^;
    Write(Log, 'Play command ', Command, ' received ... ');

    if Command <> EndOfSlice then
    begin
      if (Command and FlagMask) > 0 then
      begin
        (* block repositioning; extract block number and read new position *)
        Command := Command and BlockMask;
        Inc(ListPtr);
        PosX := ListPtr^;
        Inc(ListPtr);
        PosY := ListPtr^;
        Writeln(Log, 'overriding default position of block #', Command,
          ' to ', PosX, ', ', PosY);
        DrawBlockAt(BlockList[Command], PosX, PosY);
      end
      else
      begin
        Writeln(Log, 'default draw of block #', Command);
        DrawBlock(BlockList[Command]);
      end
    end;
    Inc(ListPtr)
  end;
  Writeln(Log, 'slice end.')
end;

procedure Play(ControlList: TFrameList; BlockList: TBlockList);
var
  ListPtr  : ^Word;
  i        : Word;
begin
  Writeln(Log, 'Starting playback');

  for i := 0 to NrSlices - 1 do
  begin
    ListPtr := ControlList[i].Content;
    if ListPtr^ <> EndOfFile then
    begin
      DrawSequence(ControlList, i);
      Delay(100)
    end
    else
      Writeln(Log, 'Semantic end of playback file detected.')
  end;

  Writeln(Log, 'Playback has finished.')
end;

begin
  Assign(Log, LogName);
  Rewrite(Log);

  (* check for file name parameter *)
  if ParamCount > 0 then
  begin
    Writeln(Log, 'Launched with first parameter ', ParamStr(1));

    Assign(VDAFile, ParamStr(1) + VDAExt);
    {$I-}
    Reset(VDAFile, 1);
    {$I+}
    if IOResult = 0 then
    begin
      Writeln(Log, 'Success opening ', ParamStr(1) + VDAExt);
      Write(Log, 'Trying to find control file ... ');

      Assign(VDXFile, ParamStr(1) + VDXExt);
      {$I-}
      Reset(VDXFile, 1);
      {$I+}
      if IOResult = 0 then
      begin
        Writeln(Log, 'success, ', ParamStr(1) + VDXExt, ' opened.');
        Writeln(Log, 'Switching into graphics mode');

        SetupMode(_320x200x8bpp);

        (* read colors *)
        Seek(VDAFile, 20);
        ReadPalette(Colors);

        (* preload animation *)
        PreloadBlocks(BlockList, NrBlocks);

        (* load animation control file *)
        LoadControl(ControlList, NrSlices);

        (* draw animation *)
        Play(ControlList, BlockList);

        (* release memory *)
        Write(Log, 'Freeing allocated memory ... ');
        BlockCount := 0;
        while BlockCount < NrBlocks do
        begin
          with BlockList[BlockCount] do
            FreeMem(Content, Longint(Width + 1) * Longint(Height + 1));
          Inc(BlockCount)
        end;

        SliceCount := 0;
        while SliceCount < NrSlices do
        begin
          with ControlList[SliceCount] do
            FreeMem(Content, DataSize);
          Inc(SliceCount)
        end;
        Writeln(Log, 'done.');

        Writeln(Log, 'Waiting for [ENTER] key to quit ...');
        Readln;

        Writeln(Log, 'Falling back to character mode');
        TextMode(co80);
        Close(VDXFile)
      end
      else
      begin
        Writeln(Log, 'failed, ', ParamStr(1) + VDXExt,
          ' could not be opened.');
        Writeln('Couldn''t open corresponding .VDX file.')
      end;
      Close(VDAFile)
    end
    else
    begin
      Writeln(Log, 'Attempt to open ', ParamStr(1) + VDAExt, ' failed');
      Writeln('Couldn''t open .VDA file.')
    end
  end
  else
  begin
    Writeln('Holmes *.VDA animation viewer');
    Writeln('(C)opyright Deniz Oezmen <deniz.oezmen@t-online.de>, 2001-2004');
    Writeln('Use at your own risk.');
    Writeln;
    Writeln('Usage: VDAVIEW4.EXE <vdafile>');
    Writeln('       Supply only the prefix, .VDA will be added.');
    Writeln('       Make sure the accompanying .VDX file is present.')
  end;
  Writeln(Log, 'Quitting application');
  Close(Log)
end.